#include <stdio.h>
#include <string.h>
#include <mpi.h>

const int MAX_STRING = 150;
const int CHUNK_SIZE = 1000;

const int WORK_TAG = 1;
const int DIE_TAG = 2;

const int DO_PRINT = 1;
const char *fileName = "/home/nverzani/HPC/file.txt";

void master();
void slave();
void countFile(FILE *file, int *numLines, int *lineLength);
int min(int a, int b);

int main(int argc, char *argv[]) {

	// TODO for next week: 
	// Look in the book to see if the master can do work too. Non blocking Recv?
	// (If I have time) Running on multiple machines? Hybrid MPI


	MPI_Init(NULL, NULL); // Initialize MPI World

	int my_rank; // Rank of specific processor
	MPI_Comm_rank(MPI_COMM_WORLD, &my_rank); // Get current processor rank

	if (my_rank == 0)
		master();
	else
		slave();
	
	MPI_Finalize();
	return 0;
}

void master() {

	int comm_sz; // Number of processors
	MPI_Comm_size(MPI_COMM_WORLD, &comm_sz); // Get number of processors

	int my_rank; // Rank of specific processor
	MPI_Comm_rank(MPI_COMM_WORLD, &my_rank); // Get current processor rank

	MPI_Status status; // Will get sender rank into here
	int count[2]; // Return results into this array

	// Open text file
	FILE *file;
	file = fopen(fileName, "r");

	// Count its length
	int numLines = 0;
	int lineLength = 0;
	countFile(file, &numLines, &lineLength);

	int sumZeros = 0;
	int sumOnes = 0;

	int currLine = 0;
	
	// Seed the slaves
	for (int i = 1; i < comm_sz; i++) {

		if (DO_PRINT)
			printf("[%d] Telling process %d to count from line %d to %d.\n", 
					my_rank, i, currLine, currLine + CHUNK_SIZE - 1);

		MPI_Send(&currLine, 1, MPI_INT, i, 0, MPI_COMM_WORLD);
		currLine = currLine + CHUNK_SIZE;
	}

	// Now, wait for tasks to finish their task, and send them a new one

	while (currLine < numLines) {

		// Wait for processors to return that they've completed a task.
		MPI_Recv(count, 2, MPI_INT, MPI_ANY_SOURCE, 0, MPI_COMM_WORLD, &status);

		// Save what it returns.
		sumZeros += count[0];
		sumOnes += count[1];

		if (DO_PRINT)
			printf("[%d] Process %d just counted %d zeros and %d ones.\n", 
					my_rank, status.MPI_SOURCE, count[0], count[1]);

		// Send it the new lines to count.

		MPI_Send(&currLine, 1, MPI_INT, status.MPI_SOURCE, WORK_TAG, MPI_COMM_WORLD);

		if (DO_PRINT)
			printf("[%d] Telling process %d to count from line %d to %d.\n", 
					my_rank, status.MPI_SOURCE, currLine, currLine + CHUNK_SIZE - 1);

		// The next one will do the next CHUNK_SIZE lines
		currLine = currLine + CHUNK_SIZE;
	}

	for (int i = 1; i < comm_sz; i++) {

		// Wait for processors to return that they've completed their last task.
		MPI_Recv(count, 2, MPI_INT, MPI_ANY_SOURCE, 0, MPI_COMM_WORLD, &status);

		if (DO_PRINT)
			printf("[%d] Process %d just counted %d zeros and %d ones.\n", 
					my_rank, status.MPI_SOURCE, count[0], count[1]);

		// Save what it returns.
		sumZeros += count[0];
		sumOnes += count[1];

		MPI_Send(0, 0, MPI_INT, status.MPI_SOURCE, DIE_TAG, MPI_COMM_WORLD);

		if (DO_PRINT)
			printf("[%d] Telling process %d to stop counting.\n", my_rank, i);
	}

	// Print what we found
	printf("\n\n **** Found %d ones and %d zeros. **** \n\n", sumOnes, sumZeros);

	fclose(file);
}

void slave() {

	int my_rank; // Rank of specific processor
	MPI_Comm_rank(MPI_COMM_WORLD, &my_rank); // Get current processor rank

	// Open text file, count its length
	FILE *file;
	file = fopen(fileName, "r");

	int numLines = 0;
	int lineLength = 0;
	countFile(file, &numLines, &lineLength);

	int startLine = 0;
	MPI_Status status;

	while (1) {

		// Receive a command from the master, and save those variables.
		MPI_Recv(&startLine, 1, MPI_INT, 0, MPI_ANY_TAG, MPI_COMM_WORLD, &status);

		// If the tag is DIE_TAG, we want to end
		if (status.MPI_TAG == DIE_TAG)
			return;

		if (DO_PRINT)
			printf("[%d] Just told to count from %d to %d.\n",
				    my_rank, startLine, startLine + CHUNK_SIZE);

		int numZeros = 0;
		int numOnes = 0;

		// This for loop is just to artifically make the program slower.
		for (int j = 0; j < 10; j++) {
			
			char buff[MAX_STRING];
			numZeros = 0;
			numOnes = 0;

			// Loop through all the lines we need to do, save them to buff
			for (int line = startLine; line < min(startLine + CHUNK_SIZE, numLines); line++) {
				
				fseek(file, line * lineLength, SEEK_SET);
				fgets(buff, MAX_STRING, file);

				// Count the number of 1's and 0's in buff
				for (int i = 0; i < lineLength; i++) {
					if (buff[i] == '1') {
						numOnes++;
					}
					else if (buff[i] == '0') {
						numZeros++;
					}
				}
			}
		}

		// Once it's counted the 1's and 0's, bundle that information up, and send it back.
		int count[2] = {numZeros, numOnes};
		MPI_Send(count, 2, MPI_INT, 0, 0, MPI_COMM_WORLD);

		if (DO_PRINT)
			printf("[%d] Telling the master it found %d ones and %d zeros from %d to %d.\n",
				    my_rank, numZeros, numOnes, startLine, min(startLine + CHUNK_SIZE, numLines));

		// Wait for the next task
	}

	fclose(file);
}

void countFile(FILE *file, int *numLines, int *lineLength) {
	
	int ch = 0;
	do
	{
		ch = fgetc(file);
		(*lineLength)++;
	} while (ch != '\n');

	// Count the number of lines in the file
	fseek(file, 0, SEEK_END);
	*numLines = ftell(file)/(*lineLength);
}

int min(int a, int b) {
	return a < b ? a : b;
}
